/* vi:set ts=4 sts=4 sw=4 et:
 *
 * VIM - Vi IMproved		by Bram Moolenaar
 *				Wasm support by rhysd <https://github.com/rhysd>
 *
 * Do ":help uganda"  in Vim to read copying and usage conditions.
 * Do ":help credits" in Vim to see a list of people who contributed.
 * See README.txt for an overview of the Vim source code.
 */
function noop() {
    /* do nothing */
}
let debug = noop;
const STATUS_NOTIFY_KEY = 1;
const STATUS_NOTIFY_RESIZE = 2;
const STATUS_REQUEST_OPEN_FILE_BUF = 3;
const STATUS_NOTIFY_OPEN_FILE_BUF_COMPLETE = 4;
const STATUS_REQUEST_CLIPBOARD_BUF = 5;
const STATUS_NOTIFY_CLIPBOARD_WRITE_COMPLETE = 6;
const STATUS_REQUEST_CMDLINE = 7;
export function checkBrowserCompatibility() {
    function notSupported(feat) {
        return `${feat} is not supported by this browser. If you're using Firefox or Safari, please enable feature flag.`;
    }
    if (typeof SharedArrayBuffer === 'undefined') {
        return notSupported('SharedArrayBuffer');
    }
    if (typeof Atomics === 'undefined') {
        return notSupported('Atomics API');
    }
    return undefined;
}
export class VimWorker {
    constructor(scriptPath, onMessage, onError) {
        this.worker = new Worker(scriptPath);
        this.worker.onmessage = this.recvMessage.bind(this);
        this.worker.onerror = this.recvError.bind(this);
        this.sharedBuffer = new Int32Array(new SharedArrayBuffer(Int32Array.BYTES_PER_ELEMENT * 128));
        this.onMessage = onMessage;
        this.onError = onError;
        this.onOneshotMessage = new Map();
        this.debug = false;
    }
    terminate() {
        this.worker.terminate();
        this.worker.onmessage = null;
        debug('Terminated worker thread. Thank you for working hard!');
    }
    sendStartMessage(msg) {
        this.worker.postMessage(msg);
        debug('Sent start message', msg);
    }
    writeOpenFileRequestEvent(name, size) {
        let idx = 1;
        this.sharedBuffer[idx++] = size;
        idx = this.encodeStringToBuffer(name, idx);
        debug('Encoded open file size event with', idx * 4, 'bytes');
        this.awakeWorkerThread(STATUS_REQUEST_OPEN_FILE_BUF);
    }
    notifyOpenFileBufComplete() {
        this.awakeWorkerThread(STATUS_NOTIFY_OPEN_FILE_BUF_COMPLETE);
    }
    notifyKeyEvent(key, keyCode, ctrl, shift, alt, meta) {
        let idx = 1;
        this.sharedBuffer[idx++] = keyCode;
        this.sharedBuffer[idx++] = +ctrl;
        this.sharedBuffer[idx++] = +shift;
        this.sharedBuffer[idx++] = +alt;
        this.sharedBuffer[idx++] = +meta;
        idx = this.encodeStringToBuffer(key, idx);
        debug('Encoded key event with', idx * 4, 'bytes');
        this.awakeWorkerThread(STATUS_NOTIFY_KEY);
        debug('Sent key event:', key, keyCode, ctrl, shift, alt, meta);
    }
    notifyResizeEvent(width, height) {
        let idx = 1;
        this.sharedBuffer[idx++] = width;
        this.sharedBuffer[idx++] = height;
        debug('Encoded resize event with', idx * 4, 'bytes');
        this.awakeWorkerThread(STATUS_NOTIFY_RESIZE);
        debug('Sent resize event:', width, height);
    }
    async requestOpenFileBuf(name, contents) {
        const size = contents.byteLength;
        let idx = 1;
        this.sharedBuffer[idx++] = size;
        idx = this.encodeStringToBuffer(name, idx);
        debug('Encoded open file size event with', idx * 4, 'bytes');
        this.awakeWorkerThread(STATUS_REQUEST_OPEN_FILE_BUF);
        const msg = (await this.waitForOneshotMessage('open-file-buf:response'));
        if (name !== msg.name) {
            // Fatal
            throw new Error(`File name mismatch from worker: '${name}' v.s. '${msg.name}'`);
        }
        if (size !== msg.buffer.byteLength) {
            // Fatal
            throw new Error(`Size of shared buffer from worker ${msg.buffer.byteLength} bytes mismatches to file contents size ${size} bytes`);
        }
        return msg.buffer;
    }
    async responseClipboardText(text, cannotSend) {
        if (cannotSend) {
            this.sharedBuffer[1] = +true;
            debug('Reading clipboard failed. Notify it to worker');
            this.awakeWorkerThread(STATUS_REQUEST_CLIPBOARD_BUF);
            return;
        }
        const encoded = new TextEncoder().encode(text);
        this.sharedBuffer[1] = +false;
        this.sharedBuffer[2] = encoded.byteLength;
        debug('Requesting', encoded.byteLength, 'bytes buffer to worker to send clipboard text:', text);
        this.awakeWorkerThread(STATUS_REQUEST_CLIPBOARD_BUF);
        const msg = (await this.waitForOneshotMessage('clipboard-buf:response'));
        new Uint8Array(msg.buffer).set(encoded);
        this.awakeWorkerThread(STATUS_NOTIFY_CLIPBOARD_WRITE_COMPLETE);
        debug('Wrote clipboard', encoded.byteLength, 'bytes text and notified to worker');
    }
    async requestCmdline(cmdline) {
        if (cmdline.length === 0) {
            throw new Error('Specified command line is empty');
        }
        const idx = this.encodeStringToBuffer(cmdline, 1);
        debug('Encoded request cmdline event with', idx * 4, 'bytes');
        this.awakeWorkerThread(STATUS_REQUEST_CMDLINE);
        const msg = (await this.waitForOneshotMessage('cmdline:response'));
        debug('Result of command', cmdline, ':', msg.success);
        if (!msg.success) {
            throw Error(`Command '${cmdline}' was invalid and not accepted by Vim`);
        }
    }
    async waitForOneshotMessage(kind) {
        return new Promise(resolve => {
            this.onOneshotMessage.set(kind, resolve);
        });
    }
    encodeStringToBuffer(s, startIdx) {
        let idx = startIdx;
        const len = s.length;
        this.sharedBuffer[idx++] = len;
        for (let i = 0; i < len; ++i) {
            this.sharedBuffer[idx++] = s.charCodeAt(i);
        }
        return idx;
    }
    awakeWorkerThread(event) {
        // TODO: Queueing request/notification to worker and wait status byte is cleared
        // Note: Non-zero means data remains not handled by worker yet.
        if (this.debug) {
            const status = Atomics.load(this.sharedBuffer, 0);
            if (status !== 0) {
                console.error('INVARIANT ERROR! Status byte must be zero cleared:', status); // eslint-disable-line no-console
            }
        }
        Atomics.store(this.sharedBuffer, 0, event);
        Atomics.notify(this.sharedBuffer, 0, 1);
        debug('Notified status event', event, 'to worker');
    }
    recvMessage(e) {
        const msg = e.data;
        // Handle oneshot communication for RPC call
        const handler = this.onOneshotMessage.get(msg.kind);
        if (handler !== undefined) {
            this.onOneshotMessage.delete(msg.kind);
            handler(msg);
            return;
        }
        // On notification
        this.onMessage(msg);
    }
    recvError(e) {
        debug('Received an error from worker:', e);
        const msg = `${e.message} (${e.filename}:${e.lineno}:${e.colno})`;
        this.onError(new Error(msg));
    }
}
export class ResizeHandler {
    constructor(domWidth, domHeight, canvas, worker) {
        this.canvas = canvas;
        this.worker = worker;
        this.elemHeight = domHeight;
        this.elemWidth = domWidth;
        const dpr = window.devicePixelRatio || 1;
        this.canvas.width = domWidth * dpr;
        this.canvas.height = domHeight * dpr;
        this.bounceTimerToken = null;
        this.onResize = this.onResize.bind(this);
    }
    onVimInit() {
        window.addEventListener('resize', this.onResize, { passive: true });
    }
    onVimExit() {
        window.removeEventListener('resize', this.onResize);
    }
    doResize() {
        const rect = this.canvas.getBoundingClientRect();
        debug('Resize Vim:', rect);
        this.elemWidth = rect.width;
        this.elemHeight = rect.height;
        const res = window.devicePixelRatio || 1;
        this.canvas.width = rect.width * res;
        this.canvas.height = rect.height * res;
        this.worker.notifyResizeEvent(rect.width, rect.height);
    }
    onResize() {
        if (this.bounceTimerToken !== null) {
            window.clearTimeout(this.bounceTimerToken);
        }
        this.bounceTimerToken = window.setTimeout(() => {
            this.bounceTimerToken = null;
            this.doResize();
        }, 500);
    }
}
// TODO: IME support
// TODO: Handle pre-edit IME state
// TODO: Follow cursor position
export class InputHandler {
    constructor(worker, input) {
        this.worker = worker;
        this.elem = input;
        // TODO: Bind compositionstart event
        // TODO: Bind compositionend event
        this.onKeydown = this.onKeydown.bind(this);
        this.onBlur = this.onBlur.bind(this);
        this.onFocus = this.onFocus.bind(this);
        this.focus();
    }
    setFont(name, size) {
        this.elem.style.fontFamily = name;
        this.elem.style.fontSize = size + 'px';
    }
    focus() {
        this.elem.focus();
    }
    onVimInit() {
        this.elem.addEventListener('keydown', this.onKeydown, { capture: true });
        this.elem.addEventListener('blur', this.onBlur);
        this.elem.addEventListener('focus', this.onFocus);
    }
    onVimExit() {
        this.elem.removeEventListener('keydown', this.onKeydown);
        this.elem.removeEventListener('blur', this.onBlur);
        this.elem.removeEventListener('focus', this.onFocus);
    }
    onKeydown(event) {
        event.preventDefault();
        event.stopPropagation();
        debug('onKeydown():', event, event.key, event.keyCode);
        let key = event.key;
        const ctrl = event.ctrlKey;
        const shift = event.shiftKey;
        const alt = event.altKey;
        const meta = event.metaKey;
        if (key.length > 1) {
            if (key === 'Unidentified' ||
                (ctrl && key === 'Control') ||
                (shift && key === 'Shift') ||
                (alt && key === 'Alt') ||
                (meta && key === 'Meta')) {
                debug('Ignore key input', key);
                return;
            }
        }
        // Note: Yen needs to be fixed to backslash
        // Note: Also check event.code since Ctrl + yen is recognized as Ctrl + | due to Chrome bug.
        // https://bugs.chromium.org/p/chromium/issues/detail?id=871650
        if (key === '\u00A5' || (!shift && key === '|' && event.code === 'IntlYen')) {
            key = '\\';
        }
        this.worker.notifyKeyEvent(key, event.keyCode, ctrl, shift, alt, meta);
    }
    onFocus() {
        debug('onFocus()');
        // TODO: Send <FocusGained> special character
    }
    onBlur(event) {
        debug('onBlur():', event);
        event.preventDefault();
        // TODO: Send <FocusLost> special character
    }
}
// Origin is at left-above.
//
//      O-------------> x
//      |
//      |
//      |
//      |
//      V
//      y
export class ScreenCanvas {
    // Note: BG color is actually unused because color information is included
    // in drawRect event arguments
    // private bgColor: string;
    constructor(worker, canvas, input) {
        this.worker = worker;
        this.canvas = canvas;
        const ctx = this.canvas.getContext('2d', { alpha: false });
        if (ctx === null) {
            throw new Error('Cannot get 2D context for <canvas>');
        }
        this.ctx = ctx;
        const rect = this.canvas.getBoundingClientRect();
        const res = window.devicePixelRatio || 1;
        this.canvas.width = rect.width * res;
        this.canvas.height = rect.height * res;
        this.canvas.addEventListener('click', this.onClick.bind(this), {
            capture: true,
            passive: true,
        });
        this.input = new InputHandler(this.worker, input);
        this.resizer = new ResizeHandler(rect.width, rect.height, canvas, worker);
        this.onAnimationFrame = this.onAnimationFrame.bind(this);
        this.queue = [];
        this.rafScheduled = false;
        this.perf = false;
    }
    onVimInit() {
        this.input.onVimInit();
        this.resizer.onVimInit();
    }
    onVimExit() {
        this.input.onVimExit();
        this.resizer.onVimExit();
    }
    draw(msg) {
        if (!this.rafScheduled) {
            window.requestAnimationFrame(this.onAnimationFrame);
            this.rafScheduled = true;
        }
        this.queue.push(msg);
    }
    focus() {
        this.input.focus();
    }
    getDomSize() {
        return {
            width: this.resizer.elemWidth,
            height: this.resizer.elemHeight,
        };
    }
    setPerf(enabled) {
        this.perf = enabled;
    }
    setColorFG(name) {
        this.fgColor = name;
    }
    setColorBG(_name) {
        // Note: BG color is actually unused because color information is included
        // in drawRect event arguments
        // this.bgColor = name;
    }
    setColorSP(name) {
        this.spColor = name;
    }
    setFont(name, size) {
        this.fontName = name;
        this.input.setFont(name, size);
    }
    drawRect(x, y, w, h, color, filled) {
        const dpr = window.devicePixelRatio || 1;
        x = Math.floor(x * dpr);
        y = Math.floor(y * dpr);
        w = Math.floor(w * dpr);
        h = Math.floor(h * dpr);
        this.ctx.fillStyle = color;
        if (filled) {
            this.ctx.fillRect(x, y, w, h);
        }
        else {
            this.ctx.rect(x, y, w, h);
        }
    }
    drawText(text, ch, lh, cw, x, y, bold, underline, undercurl, strike) {
        const dpr = window.devicePixelRatio || 1;
        ch = ch * dpr;
        lh = lh * dpr;
        cw = cw * dpr;
        x = x * dpr;
        y = y * dpr;
        let font = Math.floor(ch) + 'px ' + this.fontName;
        if (bold) {
            font = 'bold ' + font;
        }
        this.ctx.font = font;
        // Note: 'ideographic' is not available (#23)
        //   https://twitter.com/Linda_pp/status/1139373687474278400
        this.ctx.textBaseline = 'bottom';
        this.ctx.fillStyle = this.fgColor;
        const descent = (lh - ch) / 2;
        const yi = Math.floor(y + lh - descent);
        for (let i = 0; i < text.length; ++i) {
            const c = text[i];
            if (c === ' ') {
                // Note: Skip rendering whitespace
                // XXX: This optimization assumes current font renders nothing on whitespace.
                continue;
            }
            this.ctx.fillText(c, Math.floor(x + cw * i), yi);
        }
        if (underline) {
            this.ctx.strokeStyle = this.fgColor;
            this.ctx.lineWidth = 1 * dpr;
            this.ctx.setLineDash([]);
            this.ctx.beginPath();
            // Note: 3 is set with considering the width of line.
            const underlineY = Math.floor(y + lh - descent - 3 * dpr);
            this.ctx.moveTo(Math.floor(x), underlineY);
            this.ctx.lineTo(Math.floor(x + cw * text.length), underlineY);
            this.ctx.stroke();
        }
        else if (undercurl) {
            this.ctx.strokeStyle = this.spColor;
            this.ctx.lineWidth = 1 * dpr;
            const curlWidth = Math.floor(cw / 3);
            this.ctx.setLineDash([curlWidth, curlWidth]);
            this.ctx.beginPath();
            // Note: 3 is set with considering the width of line.
            const undercurlY = Math.floor(y + lh - descent - 3 * dpr);
            this.ctx.moveTo(Math.floor(x), undercurlY);
            this.ctx.lineTo(Math.floor(x + cw * text.length), undercurlY);
            this.ctx.stroke();
        }
        else if (strike) {
            this.ctx.strokeStyle = this.fgColor;
            this.ctx.lineWidth = 1 * dpr;
            this.ctx.beginPath();
            const strikeY = Math.floor(y + lh / 2);
            this.ctx.moveTo(Math.floor(x), strikeY);
            this.ctx.lineTo(Math.floor(x + cw * text.length), strikeY);
            this.ctx.stroke();
        }
    }
    invertRect(x, y, w, h) {
        const dpr = window.devicePixelRatio || 1;
        x = Math.floor(x * dpr);
        y = Math.floor(y * dpr);
        w = Math.floor(w * dpr);
        h = Math.floor(h * dpr);
        const img = this.ctx.getImageData(x, y, w, h);
        const data = img.data;
        const len = data.length;
        for (let i = 0; i < len; ++i) {
            data[i] = 255 - data[i];
            ++i;
            data[i] = 255 - data[i];
            ++i;
            data[i] = 255 - data[i];
            ++i; // Skip alpha
        }
        this.ctx.putImageData(img, x, y);
    }
    imageScroll(x, sy, dy, w, h) {
        const dpr = window.devicePixelRatio || 1;
        x = Math.floor(x * dpr);
        sy = Math.floor(sy * dpr);
        dy = Math.floor(dy * dpr);
        w = Math.floor(w * dpr);
        h = Math.floor(h * dpr);
        this.ctx.drawImage(this.canvas, x, sy, w, h, x, dy, w, h);
    }
    onClick() {
        this.input.focus();
    }
    onAnimationFrame() {
        debug('Rendering', this.queue.length, 'events on animation frame');
        this.perfMark('raf');
        for (const [method, args] of this.queue) {
            this.perfMark('draw');
            this[method].apply(this, args);
            this.perfMeasure('draw', `draw:${method}`);
        }
        this.queue.length = 0; // Clear queue
        this.rafScheduled = false;
        this.perfMeasure('raf');
    }
    perfMark(m) {
        if (this.perf) {
            performance.mark(m);
        }
    }
    perfMeasure(m, n) {
        if (this.perf) {
            performance.measure(n || m, m);
            performance.clearMarks(m);
        }
    }
}
export class VimWasm {
    constructor(opts) {
        const script = opts.workerScriptPath || './vim.js';
        this.worker = new VimWorker(script, this.onMessage.bind(this), this.onErr.bind(this));
        if ('canvas' in opts && 'input' in opts) {
            this.screen = new ScreenCanvas(this.worker, opts.canvas, opts.input);
        }
        else if ('screen' in opts) {
            this.screen = opts.screen;
        }
        else {
            throw new Error('Invalid options for VimWasm construction: ' + JSON.stringify(opts));
        }
        this.perf = false;
        this.debug = false;
        this.perfMessages = {};
        this.running = false;
        this.end = false;
    }
    start(opts) {
        if (this.running || this.end) {
            throw new Error('Cannot start Vim twice');
        }
        const o = opts || { clipboard: navigator.clipboard !== undefined };
        if (o.debug) {
            debug = console.log.bind(console, 'main:'); // eslint-disable-line no-console
            this.worker.debug = true;
        }
        this.perf = !!o.perf;
        this.debug = !!o.debug;
        this.screen.setPerf(this.perf);
        this.running = true;
        this.perfMark('init');
        const { width, height } = this.screen.getDomSize();
        const msg = {
            kind: 'start',
            buffer: this.worker.sharedBuffer,
            canvasDomWidth: width,
            canvasDomHeight: height,
            debug: this.debug,
            perf: this.perf,
            clipboard: !!o.clipboard,
            files: o.files || {},
            dirs: o.dirs || [],
            persistent: o.persistentDirs || [],
        };
        this.worker.sendStartMessage(msg);
        debug('Started with drawer', this.screen);
    }
    // Note: Sending file to Vim requires some message interactions.
    //
    // 1. Main sends FILE_REQUEST event with file size and file name to worker via shared memory buffer
    // 2. Worker waits the event with Atomics.wait() and gets the size and name
    // 3. Worker allocates a new SharedArrayBuffer with the file size
    // 4. Worker sends the buffer to main via 'file-buffer' message using postMessage()
    // 5. Main receives the message and copy file contents to the buffer
    // 6. Main sends FILE_WRITE_COMPLETE event to worker via shared memory buffer
    // 7. Worker waits the event with Atomics.wait()
    // 8. Worker reads file contents from the buffer alolocated at 3. and deletes the buffer
    // 9. Worker handles the file open
    //
    // This a bit complex interactions are necessary because postMessage() from main thread does
    // not work. Worker sleeps in Vim's main loop using Atomics.wait(). So JavaScript context in worker
    // never ends until exit() is called. It means that onmessage callback is never fired.
    async dropFile(name, contents) {
        if (!this.running) {
            throw new Error('Cannot open file since Vim is not running');
        }
        debug('Handling to open file', name, contents);
        // Get shared buffer to write file contents from worker
        const buffer = await this.worker.requestOpenFileBuf(name, contents);
        // Write file contents
        new Uint8Array(buffer).set(new Uint8Array(contents));
        // Notify worker to start processing the file contents
        this.worker.notifyOpenFileBufComplete();
        debug('Wrote file', name, 'to', contents.byteLength, 'bytes buffer and notified it to worker');
    }
    async dropFiles(files) {
        const reader = new FileReader();
        for (const file of files) {
            const [name, contents] = await this.readFile(reader, file);
            this.dropFile(name, contents);
        }
    }
    resize(pixelWidth, pixelHeight) {
        this.worker.notifyResizeEvent(pixelWidth, pixelHeight);
    }
    sendKeydown(key, keyCode, modifiers) {
        const { ctrl = false, shift = false, alt = false, meta = false } = modifiers || {};
        if (key.length > 1) {
            if (key === 'Unidentified' ||
                (ctrl && key === 'Control') ||
                (shift && key === 'Shift') ||
                (alt && key === 'Alt') ||
                (meta && key === 'Meta')) {
                debug('Ignore key input', key);
                return;
            }
        }
        this.worker.notifyKeyEvent(key, keyCode, ctrl, shift, alt, meta);
    }
    // Note: This command execution does not trigger screen redraw.
    // Please run :redraw like "1put | redraw" if updating screen is necessary.
    cmdline(cmdline) {
        return this.worker.requestCmdline(cmdline);
    }
    isRunning() {
        return this.running;
    }
    focus() {
        this.screen.focus();
    }
    async readFile(reader, file) {
        return new Promise((resolve, reject) => {
            reader.onload = f => {
                debug('Read file', file.name, 'from D&D:', f);
                resolve([file.name, reader.result]);
            };
            reader.onerror = () => {
                reader.abort();
                reject(new Error(`Error on loading file ${file}`));
            };
            reader.readAsArrayBuffer(file);
        });
    }
    onMessage(msg) {
        if (this.perf && msg.timestamp !== undefined) {
            // performance.now() is not available because time origin is different between Window and Worker
            const duration = Date.now() - msg.timestamp;
            const name = msg.kind === 'draw' ? `draw:${msg.event[0]}` : msg.kind;
            const timestamps = this.perfMessages[name];
            if (timestamps === undefined) {
                this.perfMessages[name] = [duration];
            }
            else {
                this.perfMessages[name].push(duration);
            }
        }
        switch (msg.kind) {
            case 'draw':
                this.screen.draw(msg.event);
                debug('draw event', msg.event);
                break;
            case 'read-clipboard:request':
                if (this.readClipboard) {
                    this.readClipboard()
                        .then(text => this.worker.responseClipboardText(text))
                        .catch(err => {
                        debug('Cannot read clipboard:', err);
                        return this.worker.responseClipboardText('', true);
                    });
                }
                else {
                    debug('Cannot read clipboard because VimWasm.readClipboard is not set');
                    this.worker.responseClipboardText('', true);
                }
                break;
            case 'write-clipboard':
                debug('Handle writing text', msg.text, 'to clipboard with', this.onWriteClipboard);
                if (this.onWriteClipboard) {
                    this.onWriteClipboard(msg.text);
                }
                break;
            case 'export':
                debug('Exporting file', msg.path, 'with size', msg.contents.byteLength, 'bytes with', this.onFileExport);
                if (this.onFileExport !== undefined) {
                    this.onFileExport(msg.path, msg.contents);
                }
                break;
            case 'started':
                this.screen.onVimInit();
                if (this.onVimInit) {
                    this.onVimInit();
                }
                this.perfMeasure('init');
                debug('Vim started');
                break;
            case 'exit':
                this.screen.onVimExit();
                this.printPerfs();
                this.worker.terminate();
                if (this.onVimExit) {
                    this.onVimExit(msg.status);
                }
                debug('Vim exited with status', msg.status);
                this.perf = false;
                this.debug = false;
                this.screen.setPerf(false);
                this.running = false;
                this.end = true;
                break;
            case 'error':
                debug('Vim threw an error:', msg.message);
                this.onErr(new Error(msg.message));
                this.worker.terminate();
                break;
            default:
                throw new Error(`Unexpected message from worker: ${JSON.stringify(msg)}`);
        }
    }
    onErr(err) {
        if (this.onError) {
            this.onError(err);
        }
    }
    printPerfs() {
        if (!this.perf) {
            return;
        }
        {
            const measurements = new Map();
            for (const e of performance.getEntries()) {
                const ms = measurements.get(e.name);
                if (ms === undefined) {
                    measurements.set(e.name, [e]);
                }
                else {
                    ms.push(e);
                }
            }
            const averages = {};
            const amounts = {};
            const timings = [];
            for (const [name, ms] of measurements) {
                if (ms.length === 1 && ms[0].entryType !== 'measure') {
                    timings.push(ms[0]);
                    continue;
                }
                /* eslint-disable no-console */
                console.log(`%c${name}`, 'color: green; font-size: large');
                console.table(ms, ['duration', 'startTime']);
                /* eslint-enable no-console */
                const total = ms.reduce((a, m) => a + m.duration, 0);
                averages[name] = total / ms.length;
                amounts[name] = total;
            }
            /* eslint-disable no-console */
            console.log('%cTimings (ms)', 'color: green; font-size: large');
            console.table(timings, ['name', 'entryType', 'startTime', 'duration']);
            console.log('%cAmount: Perf Mark Durations (ms)', 'color: green; font-size: large');
            console.table(amounts);
            console.log('%cAverage: Perf Mark Durations (ms)', 'color: green; font-size: large');
            console.table(averages);
            /* eslint-enable no-console */
            performance.clearMarks();
            performance.clearMeasures();
        }
        {
            const averages = {};
            for (const name of Object.keys(this.perfMessages)) {
                const durations = this.perfMessages[name];
                const total = durations.reduce((a, d) => a + d, 0);
                averages[name] = total / durations.length;
            }
            // Note: Amounts of durations of inter-thread messages don't make sense since messaging is asynchronous. Multiple messages are sent and processed
            /* eslint-disable no-console */
            console.log('%cAverage: Inter-thread Messages Duration (ms)', 'color: green; font-size: large');
            console.table(averages);
            /* eslint-enable no-console */
            this.perfMessages = {};
        }
    }
    perfMark(m) {
        if (this.perf) {
            performance.mark(m);
        }
    }
    perfMeasure(m) {
        if (this.perf) {
            performance.measure(m, m);
            performance.clearMarks(m);
        }
    }
}
//# sourceMappingURL=vimwasm.js.map